module Feature.Sharing.SQLite where

import           Core.Configuration
import           Core.SQLite
import           Core.Types
import           Feature.Generation.Passwords      (Password)

import           Control.Monad.Trans               (liftIO)
import           Data.List.Split                   (splitOn)
import           Data.Maybe                        (listToMaybe, fromMaybe, Maybe(Just))
import           Data.Time.Clock.POSIX             (getPOSIXTime)
import           Database.SQLite.Simple

import qualified Crypto.Saltine.Core.SecretBox     as Box
import qualified Crypto.Saltine.Class              as CL 
import qualified Data.ByteString.Base64            as B64
import qualified Data.ByteString.Char8             as BSC8
import qualified Data.ByteString                   as B
import qualified Data.Text                         as T
import qualified Data.Text.Encoding                as ET
import qualified Data.Text.Lazy                    as LT

-- Look up a secret based on the "link" attribute
findByLink :: String -> PurrAction (Maybe T.Text)
findByLink link = do
  -- Get the encryption key from the filesystem as a ByteString
  key   <- liftIO encKey
  -- Start up a connection to the SQLite database
  conn  <- liftIO $ open dbPath
  -- Constant containing the results of a query looking for the "link" attribute
  res   <- liftIO $ 
    query conn "SELECT * from pws WHERE link = ?" 
    (Only (last $ splitOn "/" link))
  -- Close the SQLite database connection
  liftIO $ close conn
  -- Pass the encryption key and [SecretEntry] to be unencrypted
  readEncryptedSecret key res

readEncryptedSecret :: B.ByteString -> [SecretEntry] -> PurrAction (Maybe T.Text)
readEncryptedSecret key [] = return Nothing
readEncryptedSecret key (secret:_) = do
  -- Increment the number of views on the secret in the database by one
  liftIO $ incViews secret dbPath
  -- Delete the secret if it's expired
  delete <- liftIO $ deleteExpiredSecret secret dbPath
  if (delete) 
    -- Don't return the secret if it's expired
    then return Nothing
    -- Otherwise, try to decrypt and return it
    else return (ET.decodeLatin1 <$> 
      (decryptSecret key (nonce secret) $ decodeSecret secret))
  where
    incViews :: SecretEntry -> String -> IO ()
    incViews secret dbPath = do
      conn <- open dbPath
      execute conn
        "UPDATE pws SET views = views + 1 WHERE link = ?" (Only (link secret))
      close conn

deleteExpiredSecret :: SecretEntry -> String -> IO Bool
deleteExpiredSecret sec dbPath = do
  -- Get the current Unix Epoch time in seconds
  time <- epochTime
  -- Compare the current time against the secret's initial insertion and lifetime
  if ((date sec) + ((life sec) * 86400) < time) || (views sec >= maxViews sec)
    -- Delete the secret if it's expired and return True for the caller
    then deleteSec sec dbPath
    -- Only return False if the secret is not expired
    else return False
  where
    deleteSec :: SecretEntry -> String -> IO Bool
    deleteSec sec dbPath = do
      conn  <- open dbPath
      execute conn
        "DELETE FROM pws WHERE link = ?" (Only (link sec))
      close conn
      return True

insertNewSecret :: T.Text -> Integer -> T.Text -> Integer -> PurrAction ()
insertNewSecret sec life link maxViews = do
  key    <- liftIO encKey
  -- Create a new nonce to associate with the secret's encryption
  nonce  <- liftIO Box.newNonce
  {- Encrypt the secret; this is a pure function because we seeded RNG at program
     initialization with "sodiumInit" -}
  let encSec = encryptSecret key sec nonce
  conn   <- liftIO $ open dbPath
  -- Get the current time to timestamp the secret's initial insertion
  time   <- liftIO epochTime
  -- Save the secret to the database
  liftIO $ execute conn
    "INSERT INTO pws (link, secret, nonce, date, life, views, maxViews) VALUES (?, ?, ?, ?, ?, ?, ?)"
      (SecretEntry link (encodeSecret encSec) (CL.encode nonce) time life 0 maxViews)
  liftIO $ close conn

encodeSecret :: B.ByteString -> T.Text
encodeSecret b = ET.decodeUtf8 $ B64.encode b

decodeSecret :: SecretEntry -> B.ByteString
decodeSecret s = B64.decodeLenient $ ET.encodeUtf8 (secret s)

encryptSecret :: B.ByteString -> T.Text -> Box.Nonce ->  B.ByteString
encryptSecret k s n = do 
  case (CL.decode k) of
    (Just key) -> Box.secretbox key n (ET.encodeUtf8 s)
    Nothing -> error "fail"

decryptSecret :: B.ByteString -> B.ByteString -> B.ByteString -> Maybe B.ByteString
decryptSecret k n b = do 
  case (CL.decode k) of 
    (Just key) -> case (CL.decode n) of
      (Just nonce) -> Box.secretboxOpen key nonce b
      {- There's no sensible way to fail gracefully if our nonce or secret
         key can't be decoded. Throw an error so the instance admin can 
         investigate -}
      Nothing -> error "Failed to decode nonce"
    Nothing -> error "Failed to decode secret key"

-- Helper function to provide the current Epoch Time in seconds
epochTime :: IO Integer
epochTime = fmap round getPOSIXTime
